﻿//Josip Medved <jmedved@jmedved.com> http://www.jmedved.com

//2008-04-20: Initial version.


using System;
using System.ComponentModel;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace Medo.Windows.Forms {

	/// <summary>
	/// Form that includes functions WndProc.
	/// </summary>
	[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "Wnd", Justification = "This naming is used within Microsoft classes also.")]
	public class WndProcForm : System.Windows.Forms.Form {

		private IntPtr _wndProcOld = IntPtr.Zero;
		private WndProcDelegate _wndProcNew;

		/// <summary>
		/// Releases the unmanaged resources used by the System.Windows.Forms.Control and its child controls and optionally releases the managed resources.
		/// </summary>
		/// <param name="disposing">True to release both managed and unmanaged resources; false to release only unmanaged resources.</param>
		protected override void Dispose(bool disposing) {
			Detach();
			base.Dispose(disposing);
		}

		/// <summary>
		/// Creates new instance.
		/// </summary>
		public WndProcForm() : base() { }

		/// <summary>
		/// Raises the System.Windows.Forms.Control.HandleCreated event.
		/// </summary>
		/// <param name="e">An System.EventArgs that contains the event data.</param>
		protected override void OnHandleCreated(System.EventArgs e) {
			base.OnHandleCreated(e);
			Attach();
		}

		/// <summary>
		/// Raises the System.Windows.Forms.Control.HandleCreated event.
		/// </summary>
		/// <param name="e">An System.EventArgs that contains the event data.</param>
		protected override void OnHandleDestroyed(System.EventArgs e) {
			Detach();
			base.OnHandleDestroyed(e);
		}


		private void Attach() {
			this._wndProcNew = new WndProcDelegate(WndProc);
			this._wndProcOld = NativeMethods.GetWindowLong(this.Handle, NativeMethods.GWL_WNDPROC);
			if (_wndProcOld == IntPtr.Zero) throw new Win32Exception(0, "Cannot retrieve old window procedure.");
			IntPtr ret = NativeMethods.SetWindowLong(this.Handle, NativeMethods.GWL_WNDPROC, _wndProcNew);
			if (ret == IntPtr.Zero) throw new Win32Exception(0, "Cannot set new windows procedure.");
		}


		private void Detach() {
			if (this._wndProcOld != IntPtr.Zero) { //Dispose can be called multiple times.
				IntPtr ret = NativeMethods.SetWindowLong(this.Handle, NativeMethods.GWL_WNDPROC, _wndProcOld);
				if (ret == IntPtr.Zero) throw new Win32Exception(0, "Cannot set old window procedure.");
				this._wndProcOld = IntPtr.Zero;
			}
		}

		/// <summary>
		/// Windows procedure handler.
		/// </summary>
		/// <param name="hWnd">Window handle of the message.</param>
		/// <param name="msg">ID number for the message. It is defined as signed integer to be CLR compatible but it should be treated as unsigned value.</param>
		/// <param name="wparam">WParam field of the message.</param>
		/// <param name="lparam">LParam field of the message.</param>
		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "Wnd", Justification = "This naming is used within Microsoft classes also.")]
		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "wparam", Justification = "This naming is used within Microsoft classes also.")]
		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "lparam", Justification = "This naming is used within Microsoft classes also.")]
		[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "h", Justification = "This naming is used within Microsoft classes also.")]
		protected virtual void WndProc(IntPtr hWnd, Int32 msg, IntPtr wparam, IntPtr lparam) {
			NativeMethods.CallWindowProc(_wndProcOld, hWnd, unchecked((UInt32)msg), wparam, lparam);
		}

		private delegate void WndProcDelegate(IntPtr hWnd, Int32 Msg, IntPtr wParam, IntPtr lParam);

		private static class NativeMethods {

			internal const int GWL_WNDPROC = -4;


			internal static IntPtr GetWindowLong(IntPtr hWnd, Int32 nIndex) {
				if (Environment.OSVersion.Platform == PlatformID.WinCE) {
					return WinCEGetWindowLong(hWnd, nIndex);
				} else {
					if (IntPtr.Size == 4) {
						return Win32GetWindowLong(hWnd, nIndex);
					} else {
						return Win64GetWindowLongPtr(hWnd, nIndex);
					}
				}
			}

			internal static IntPtr SetWindowLong(IntPtr hWnd, int nIndex, WndProcDelegate dwNewLong) {
				if (Environment.OSVersion.Platform == PlatformID.WinCE) {
					return WinCESetWindowLong(hWnd, nIndex, dwNewLong);
				} else {
					if (IntPtr.Size == 4) {
						return Win32SetWindowLong(hWnd, nIndex, dwNewLong);
					} else {
						return Win64SetWindowLongPtr(hWnd, nIndex, dwNewLong);
					}
				}
			}

			internal static IntPtr SetWindowLong(IntPtr hWnd, int nIndex, IntPtr dwNewLong) {
				if (Environment.OSVersion.Platform == PlatformID.WinCE) {
					return WinCESetWindowLong(hWnd, nIndex, dwNewLong);
				} else {
					if (IntPtr.Size == 4) {
						return Win32SetWindowLong(hWnd, nIndex, dwNewLong);
					} else {
						return Win64SetWindowLongPtr(hWnd, nIndex, dwNewLong);
					}
				}
			}

			internal static IntPtr CallWindowProc(IntPtr lpPrevWndFunc, IntPtr hWnd, uint Msg, IntPtr wParam, IntPtr lParam) {
				if (Environment.OSVersion.Platform == PlatformID.WinCE) {
					return WinCECallWindowProc(lpPrevWndFunc, hWnd, Msg, wParam, lParam);
				} else {
					return Win32CallWindowProc(lpPrevWndFunc, hWnd, Msg, wParam, lParam);
				}
			}


			[DllImport("coredll.dll", EntryPoint = "GetWindowLong", CharSet = CharSet.Unicode, SetLastError = true)]
			internal static extern IntPtr WinCEGetWindowLong(IntPtr hWnd, Int32 nIndex);

			[DllImport("coredll.dll", EntryPoint = "SetWindowLong", CharSet = CharSet.Unicode, SetLastError = true)]
			internal static extern IntPtr WinCESetWindowLong(IntPtr hWnd, Int32 nIndex, WndProcDelegate dwNewLong);

			[DllImport("coredll.dll", EntryPoint = "SetWindowLong", CharSet = CharSet.Unicode, SetLastError = true)]
			internal static extern IntPtr WinCESetWindowLong(IntPtr hWnd, Int32 nIndex, IntPtr dwNewLong);

			[DllImport("coredll.dll", EntryPoint = "CallWindowProc", CharSet = CharSet.Unicode, SetLastError = true)]
			internal static extern IntPtr WinCECallWindowProc(IntPtr lpPrevWndFunc, IntPtr hWnd, UInt32 Msg, IntPtr wParam, IntPtr lParam);


			[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Portability", "CA1901:PInvokeDeclarationsShouldBePortable", MessageId = "return", Justification = "Portability is handled in function before call with IntPrt.Size check.")]
			[DllImport("user32.dll", EntryPoint = "GetWindowLong", CharSet = CharSet.Unicode, SetLastError = true)]
			internal static extern IntPtr Win32GetWindowLong(IntPtr hWnd, Int32 nIndex);

			[DllImport("user32.dll", EntryPoint = "GetWindowLongPtr", CharSet = CharSet.Unicode, SetLastError = true)]
			[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Interoperability", "CA1400:PInvokeEntryPointsShouldExist", Justification = "Function does exist on 64-bit platforms.")]
			internal static extern IntPtr Win64GetWindowLongPtr(IntPtr hWnd, Int32 nIndex);

			[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Portability", "CA1901:PInvokeDeclarationsShouldBePortable", MessageId = "return", Justification = "Portability is handled in function before call with IntPrt.Size check.")]
			[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Portability", "CA1901:PInvokeDeclarationsShouldBePortable", MessageId = "2", Justification = "Portability is handled in function before call with IntPrt.Size check.")]
			[DllImport("user32.dll", EntryPoint = "SetWindowLong", CharSet = CharSet.Unicode, SetLastError = true)]
			internal static extern IntPtr Win32SetWindowLong(IntPtr hWnd, Int32 nIndex, WndProcDelegate dwNewLong);

			[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Portability", "CA1901:PInvokeDeclarationsShouldBePortable", MessageId = "return", Justification = "Portability is handled in function before call with IntPrt.Size check.")]
			[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Portability", "CA1901:PInvokeDeclarationsShouldBePortable", MessageId = "2", Justification = "Portability is handled in function before call with IntPrt.Size check.")]
			[DllImport("user32.dll", EntryPoint = "SetWindowLong", CharSet = CharSet.Unicode, SetLastError = true)]
			internal static extern IntPtr Win32SetWindowLong(IntPtr hWnd, Int32 nIndex, IntPtr dwNewLong);

			[DllImport("user32.dll", EntryPoint = "SetWindowLongPtr", CharSet = CharSet.Unicode, SetLastError = true)]
			[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Interoperability", "CA1400:PInvokeEntryPointsShouldExist", Justification = "Function does exist on 64-bit platforms.")]
			internal static extern IntPtr Win64SetWindowLongPtr(IntPtr hWnd, Int32 nIndex, WndProcDelegate dwNewLong);

			[DllImport("user32.dll", EntryPoint = "SetWindowLongPtr", CharSet = CharSet.Unicode, SetLastError = true)]
			[System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Interoperability", "CA1400:PInvokeEntryPointsShouldExist", Justification = "Function does exist on 64-bit platforms.")]
			internal static extern IntPtr Win64SetWindowLongPtr(IntPtr hWnd, Int32 nIndex, IntPtr dwNewLong);

			[DllImport("user32.dll", EntryPoint = "CallWindowProc", CharSet = CharSet.Unicode, SetLastError = true)]
			internal static extern IntPtr Win32CallWindowProc(IntPtr lpPrevWndFunc, IntPtr hWnd, UInt32 Msg, IntPtr wParam, IntPtr lParam);

		}

	}

}